package de.meinefirma.meinprojekt.client;

import java.net.URL;
import java.text.DecimalFormat;
import java.util.Random;
import javax.ws.rs.client.*;
import javax.xml.namespace.QName;
import javax.xml.ws.*;
import de.meinefirma.meinprojekt.buecher.*;
import de.meinefirma.meinprojekt.dao.BuecherUtil;

/** Performance-Testclient sowohl fuer den REST- als auch fuer den SOAP-Webservice */
public class BuecherRestAndSoapPerfTestClient
{
   public static void main( final String[] args ) throws Exception
   {
      int    anzahlBuecher = 100;
      String urlBasis      = "http://localhost:8080/JaxRsBuecherverwaltung";
      switch( Math.min( 2, args.length ) ) {
         case 2: urlBasis      = args[1];                     // $FALL-THROUGH$
         case 1: anzahlBuecher = Integer.parseInt( args[0] ); // $FALL-THROUGH$
         default:
      }
      // SOAP- und REST-SUT (System under Test):
      SoapSut soapSut = new SoapSut();
      test( soapSut,       "SOAP", urlBasis + "/ws/BuecherSoapService", anzahlBuecher, true );
      soapSut.loescheAlleBuecher();
      test( new RestSut(), "REST", urlBasis + "/rest/Artikel/Buecher",  anzahlBuecher, true );
   }

   public static BuecherTO test( SutIntf sut, String testName, String url, int anzahlBuecher, boolean trace ) throws Exception
   {
      System.out.println( "\n" + testName + ": " + url + "\n" );
      Long[] ids = new Long[anzahlBuecher];
      sut.initialize( url );
      System.gc();

      // Anlage von Buechern:
      if( trace ) System.out.println( "\n" + testName + ": Starte Anlage von " + anzahlBuecher + " Buechern" );
      long startZeit = System.nanoTime();
      for( int i = 0; i < anzahlBuecher; i++ ) {
         BuchDO bu = BuecherUtil.erzeugeBuchDO( null, "" + (1000000000L + i), "Buch " + i, new Double( i ) );
         BuecherTO bueTO = sut.createBuch( bu );
         ids[i] = bueTO.getResults().get( 0 ).getId();
      }
      String s1 = "\nAnlage von " + anzahlBuecher + " Buechern dauert: " + ermittleDauer( startZeit );

      // Auslesen von Buechern in einzelnen Lesevorgaengen:
      if( trace ) System.out.println( testName + ": Starte einzelnes Auslesen" );
      startZeit = System.nanoTime();
      for( int i = 0; i < anzahlBuecher; i++ ) {
         Long id = ids[(new Random()).nextInt( anzahlBuecher )];
         BuecherTO bueTO = sut.getBuchById( id );
         if( bueTO == null || bueTO.getResults() == null || bueTO.getResults().size() != 1 ) {
            throw new RuntimeException( "Fehler beim Auslesen des Buches mit der ID " + id );
         }
      }
      String s2 = "\nEinzelnes Auslesen von " + anzahlBuecher + " Buechern dauert: " + ermittleDauer( startZeit );

      // Auslesen aller Buecher in einem Lesevorgang:
      if( trace ) System.out.println( testName + ": Starte gemeinsames Auslesen" );
      startZeit = System.nanoTime();
      BuecherTO bueTO = sut.findeBuecher( new BuchDO() );
      String s3 = "\nGemeinsames Auslesen von " + bueTO.getResults().size() + " Buechern dauert: " + ermittleDauer( startZeit );

      // Ausgabe:
      System.out.println( "\n" + testName + ": " + s1 + s2 + s3 + "\n" );
      return bueTO;
   }

   static String zeigeErgebnis( BuecherTO bueTO )
   {
      StringBuffer sb = new StringBuffer();
      sb.append( "\n" + bueTO.getMessage() + "\n" );
      for( BuchDO bu : bueTO.getResults() )
         sb.append( "    Buch (ID=" + bu.getId() + ", ISBN=" + bu.getIsbn() + ", Titel=" + bu.getTitel() + ", Preis=" + bu.getPreis() + ")\n" );
      sb.append( "    Returncode " + bueTO.getReturncode() + "\n" );
      return sb.toString();
   }

   static String ermittleDauer( long startZeitNanoSek )
   {
      long dauerMs = (System.nanoTime() - startZeitNanoSek) / 1000 / 1000;
      if( dauerMs < 1000 ) return "" + dauerMs + " ms";
      return (new DecimalFormat( "#,##0.00" )).format( dauerMs / 1000. ) + " s";
   }
}

// REST-Implementierung des "System under Test":
class RestSut implements SutIntf
{
   WebTarget webTarget;

   @Override public void initialize( String url ) throws Exception
   {
      Client c = ClientBuilder.newClient();
      webTarget = c.target( url );
      webTarget.request().get( BuecherTO.class );
      Thread.sleep( 1000 );
   }

   @Override public BuecherTO createBuch( BuchDO bu ) throws Exception
   {
      return webTarget.request( "text/xml; charset=utf-8" ).post( Entity.entity( bu, "text/xml; charset=utf-8" ), BuecherTO.class );
   }

   @Override public BuecherTO getBuchById( Long id )
   {
      return webTarget.path( "" + id ).request().get( BuecherTO.class );
   }

   @Override public BuecherTO findeBuecher( BuchDO bu )
   {
      return webTarget.request().get( BuecherTO.class );
   }
}

// SOAP-Implementierung des "System under Test":
class SoapSut implements SutIntf
{
   BuecherSoapServiceIntf buecherService;

   @Override public void initialize( String url ) throws Exception
   {
      // Zugriff auf den Webservice vorbereiten:
      if( url.startsWith( "direkt" ) ) {
         buecherService = new BuecherSoapServiceImpl();
      } else {
         Service service = null;
         int timeoutSekunden = 20;
         while( service == null ) {
            try {
               service = Service.create(
                     new URL( url + "?wsdl" ),
                     new QName( "http://buecher.meinprojekt.meinefirma.de/", "BuecherSoapServiceImplService" ) );
            } catch( WebServiceException ex ) {
               if( timeoutSekunden-- <= 0 ) throw ex;
               Thread.sleep( 1000 );
            }
         }
         buecherService = service.getPort( BuecherSoapServiceIntf.class );
      }
   }

   @Override public BuecherTO createBuch( BuchDO bu ) throws Exception
   {
      return buecherService.createBuch( bu );
   }

   @Override public BuecherTO getBuchById( Long id )
   {
      return buecherService.getBuchById( id );
   }

   @Override public BuecherTO findeBuecher( BuchDO bu )
   {
      return buecherService.findeBuecher( bu );
   }

   public void loescheAlleBuecher()
   {
      buecherService.loescheAlleBuecher();
   }
}

// Interface fuer "System under Test":
interface SutIntf
{
   void initialize( String url ) throws Exception;
   BuecherTO createBuch( BuchDO bu ) throws Exception;
   BuecherTO getBuchById( Long id );
   BuecherTO findeBuecher( BuchDO bu );
}
