package restclient;

import javax.json.*;
import java.io.*;
import java.text.DecimalFormat;
import java.util.*;

/**
 * JSON-REST-Client zur Berechnung des RCI (Rules Compliance Index) aller Projekte fuer SonarQube 6.3,
 * {@link "https://docs.sonarqube.org/display/DEV/Web+API"}
 * {@link "http://localhost:9000/web_api"}
 * {@link "http://localhost:9000/api/components/search?qualifiers=TRK"}
 * {@link "http://localhost:9000/api/measures/component?metricKeys=ncloc&componentKey={projName}"}
 * {@link "http://localhost:9000/api/issues/search?statuses=OPEN&severities=BLOCKER&facets=severities&facetMode=count&projectKeys={projName}"}
 */
public class SonarQube63Rci
{
   public static void main( String[] args )
   {
      System.out.println(
            "\n\n---------------------------------------------------------------------------------------------------------------\n" +
            "Berechnung des RCI (Rules Compliance Index) aller Projekte fuer SonarQube 6.3,\n" +
            "1. Parameter: SonarQube-Basis-URL, 2. Parameter: Severity-RCI-Wichtung, z.B.:\n" +
            "java -cp target\\RestClient-jar-with-dependencies.jar restclient.SonarQube63Rci http://localhost:9000 10/5/3/1/0\n" +
            "---------------------------------------------------------------------------------------------------------------\n" );

      String basisUrl = "http://localhost:9000";
      int[] rciWeights = new int[] { 10, 5, 3, 1, 0 };

      if( args != null && args.length > 0 && args[0].length() > 0 ) {
         basisUrl = args[0];
      }
      if( args != null && args.length > 1 && args[1].length() > 0 ) {
         String ss[] = args[1].replace( '/', ',' ).split( "," );
         if( ss.length == rciWeights.length ) {
            for( int i = 0; i < rciWeights.length; i++ ) {
               rciWeights[i] = Integer.parseInt( ss[i] );
            }
         }
      }

      System.out.println( "SonarQube-Basis-URL:   " + basisUrl );
      System.out.println( "Severity-RCI-Wichtung: " + rciWeights[0] + "/" + rciWeights[1] + "/" + rciWeights[2] + "/" +
                                                      rciWeights[3] + "/" + rciWeights[4] + " (Blocker/Critical/Major/Minor/Info)" );

      List<SonarQubeProjektInfos> projekteInfos = findeProjekteNamen( basisUrl );
      ermittleLinesOfCodeAnzahl( basisUrl, projekteInfos );
      ermittleViolations( basisUrl, rciWeights, projekteInfos );

      System.out.println( "Anzahl der Projekte:   " + projekteInfos.size() + "\n" );

      ProjektInfosComparator comp = new ProjektInfosComparator();
      comp.sortcolumn = "projName1";

      ausgabeAufKonsole( projekteInfos );
      speichereInCsv( projekteInfos, "SonarQube-RCI.csv" );
      speichereInHtml( projekteInfos, "SonarQube-RCI-sortName.html" );

      comp.sortcolumn = "rci";
      projekteInfos.sort( comp );

      speichereInHtml( projekteInfos, "SonarQube-RCI-sortRci.html" );
      System.out.println( "" );
   }

   private static List<SonarQubeProjektInfos> findeProjekteNamen( String basisUrl )
   {
      List<SonarQubeProjektInfos> projekteInfos = new ArrayList<>();
      JsonObject jsonObj = JsonObjectFromUrlUtil.getJsonObjectFromUrl( basisUrl + "/api/components/search?qualifiers=TRK" );
      JsonArray components = jsonObj.getJsonArray( "components" );
      for( JsonObject comp : components.getValuesAs( JsonObject.class ) ) {
         SonarQubeProjektInfos pi = new SonarQubeProjektInfos();
         pi.projName2 = comp.getString( "key" );
         pi.projName1 = pi.projName2;
         int i = pi.projName1.indexOf( ':' );
         if( i > 0 ) {
            pi.projName1 = ( pi.projName1.startsWith( "schnittstellen" ) ) ? pi.projName1.substring( 0, i ) : pi.projName1.substring( ++i );
         }
         projekteInfos.add( pi );
      }
      return projekteInfos;
   }

   private static void ermittleLinesOfCodeAnzahl( String basisUrl, List<SonarQubeProjektInfos> projekteInfos )
   {
      for( SonarQubeProjektInfos pi : projekteInfos ) {
         JsonObject jsonObj = JsonObjectFromUrlUtil.getJsonObjectFromUrl(
               basisUrl + "/api/measures/component?metricKeys=ncloc&componentKey=" + pi.projName2 );
         JsonArray measures = jsonObj.getJsonObject( "component" ).getJsonArray( "measures" );
         for( JsonObject m : measures.getValuesAs( JsonObject.class ) ) {
            if( "ncloc".equalsIgnoreCase( m.getString( "metric" ) ) ) {
               try {
                  pi.linesOfCode += Integer.parseInt( m.getString( "value" ) );
               } catch( NumberFormatException e ) {
                  System.out.println( "Lines of Code konnte nicht geparsed werden aus: " + m.getString( "value" ) );
               }
            }
         }
      }
   }

   private static void ermittleViolations( String basisUrl, int[] rciWeights, List<SonarQubeProjektInfos> projekteInfos )
   {
      for( SonarQubeProjektInfos pi : projekteInfos ) {
         JsonObject jsonObj = JsonObjectFromUrlUtil.getJsonObjectFromUrl( basisUrl +
                     "/api/issues/search?statuses=OPEN&severities=BLOCKER&facets=severities&facetMode=count&projectKeys=" +
                     pi.projName2 );
         JsonArray facets = jsonObj.getJsonArray( "facets" );
         for( JsonObject violationFacets : facets.getValuesAs( JsonObject.class ) ) {
            JsonArray violationValues = violationFacets.getJsonArray( "values" );
            for( JsonObject viol : violationValues.getValuesAs( JsonObject.class ) ) {
               String val = viol.getString( "val" );
               int    cnt = viol.getInt( "count" );
               if( "BLOCKER".equalsIgnoreCase(  val ) ) { pi.blocker  = cnt; }
               if( "CRITICAL".equalsIgnoreCase( val ) ) { pi.critical = cnt; }
               if( "MAJOR".equalsIgnoreCase(    val ) ) { pi.major    = cnt; }
               if( "MINOR".equalsIgnoreCase(    val ) ) { pi.minor    = cnt; }
               if( "INFO".equalsIgnoreCase(     val ) ) { pi.info     = cnt; }
               pi.rci = berechneRci(  pi.linesOfCode, rciWeights,  pi.blocker,  pi.critical,  pi.major,  pi.minor,  pi.info );
            }
         }
      }
   }

   private static void ausgabeAufKonsole(List<SonarQubeProjektInfos> projekteInfos )
   {
      for( SonarQubeProjektInfos pi : projekteInfos ) {
         System.out.printf( pi.projName1 + ": Loc: " + pi.linesOfCode + ", Blocker: "  + pi.blocker + ", Critical: " + pi.critical +
               ", Major: " + pi.major + ", Minor: " + pi.minor + ", Info: " + pi.info + ", --> RCI: %.1f\n", pi.rci );
      }
   }

   private static void speichereInHtml(List<SonarQubeProjektInfos> projekteInfos, String htmlFilename )
   {
      try( BufferedWriter out = new BufferedWriter( new OutputStreamWriter( new FileOutputStream( htmlFilename ), "UTF-8") ) ) {
         out.write( "<html><body><table border=0 cellspacing=8 cellpadding=0>\n" );
         out.write( "<tr bgcolor='#E7EAEA'><th>Projekt</th><th>RCI</th><th>LOC</th>" +
               "<th>Blocker</th><th>Critical</th><th>Major</th><th>Minor</th><th>Info</th></tr>\n" );
         for( SonarQubeProjektInfos pi : projekteInfos ) {
            String rci = String.format( "%.1f", pi.rci );
            out.write( "<tr><td>" + pi.projName1 + "</td><td>" + rci + "</td><td>" + pi.linesOfCode + "</td>" +
                  "<td>" + pi.blocker + "</td><td>" + pi.critical + "</td><td>" + pi.major + "</td><td>" + pi.minor + "</td><td>" + pi.info + "</td></tr>\n" );
         }
         out.write( "</table></body></html>\n" );
      } catch( IOException e ) {
         e.printStackTrace();
      }
   }

   private static void speichereInCsv(List<SonarQubeProjektInfos> projekteInfos, String csvFilename )
   {
      try( BufferedWriter out = new BufferedWriter( new OutputStreamWriter( new FileOutputStream( csvFilename ), "UTF-8") ) ) {
         out.write( "Projekt;RCI;LOC;Blocker;Critical;Major;Minor;Info;\n" );
         for( SonarQubeProjektInfos pi : projekteInfos ) {
            DecimalFormat df = (DecimalFormat) DecimalFormat.getInstance( Locale.US );
            df.applyPattern( "0.0" );
            String rci = df.format( pi.rci );
            out.write( pi.projName1 + ";" + rci + ";" + pi.linesOfCode + ";" +
                  pi.blocker + ";" + pi.critical + ";" + pi.major + ";" + pi.minor + ";" + pi.info + ";\n" );
         }
      } catch( IOException e ) {
         e.printStackTrace();
      }
   }

   private static double berechneRci( int linesOfCode, int[] rciWeights, int blocker, int critical, int major, int minor, int info )
   {
      if( linesOfCode <= 0 ) { return 0.0; }
      int issueWeight = rciWeights[0] * blocker + rciWeights[1] * critical + rciWeights[2] * major + rciWeights[3] * minor + rciWeights[4] * info;
      return Math.max( 100. - ((double) issueWeight / linesOfCode) * 100., 0 );
   }

   static class SonarQubeProjektInfos
   {
      String projName1;
      String projName2;
      int linesOfCode;
      int blocker;
      int critical;
      int major;
      int minor;
      int info;
      double rci;
   }

   static class ProjektInfosComparator implements Comparator<SonarQubeProjektInfos>
   {
      String sortcolumn;

      @Override
      public int compare( SonarQubeProjektInfos pi1, SonarQubeProjektInfos pi2 ) {
         if( "rci".equalsIgnoreCase( sortcolumn ) ) {
            return Double.compare( pi1.rci, pi2.rci );
         } else {
            return pi1.projName1.compareTo( pi2.projName1 );
         }
      }
   }
}
